#!/usr/bin/python
import json
import os
import subprocess
import tempfile
import argparse

'''
The purpose of this script is to identify issues around bucket overflow / roci
imbalance / unmapped rocis and recommend which nodes might be problematic.
TSEs can decide to remove these nodes to recover the system

The information the script collects:
1. cluster / enode / devices information
2. ecs roc map
3. roci mapper counter of rocis per container
4. bucket overflow counters
5. traces related to bucket overflow
'''

#TODO: Make the script run faster by accessing the nodes in parallel.

SSH_OPTS = ['-q', '-oStrictHostKeyChecking=no']
EDEB_PORT = 8911
ROCIS_PER_CONTAINER_COUNTER_ID = 'rocis_count'
BUCKET_OVERFLOW_COUNTER_ID = 'direct.pdbl.save.bucket_'
MAX_ROCIS_COUNT_PER_CTR = 32
MAX_ROCI_IMBALANCE_BETWEEN_CTRS = 4


class ENode(object):
    def __init__(self, data_ip, devices):
        self.data_ip = data_ip
        self.devices = devices

    def __repr__(self):
        return 'ENode(data_ip={}, devices={})'.format(self.data_ip, repr(self.devices))

def cmd_output(args):
    print(('> {}'.format(' '.join(args))))
    return subprocess.check_output(args, shell=False)

def cleanup_proc(proc):
    if proc.poll() is None:
        try:
            proc.kill()
        except OSError:
            pass # assume already killed

class ProcessException(Exception):
    pass

def cmd_redirect_into(out_file, args):
    print(('> {} (output in: {})'.format(' '.join(args), out_file.name)))
    proc = subprocess.Popen(args,
                            shell=False,
                            stdout=out_file.fileno(),
                            stderr=out_file.fileno())
    try:
        proc.wait()
    finally:
        cleanup_proc(proc)
    if proc.poll() != 0:
        raise ProcessException('failed: {} - output redirect to: {}'.format(' '.join(args), out_file.name))

def get_enodes():
    enodes_json = cmd_output(['elfs-cli', '--json', 'enode', 'list'])
    enodes_dicts = json.loads(enodes_json)
    enodes = []
    for enode_dict in enodes_dicts:
        devices = [device['device_path'] for device in enode_dict['devices']]
        enode = ENode(enode_dict['data_ip'], devices)
        enodes.append(enode)
    return enodes

class PerCtr(object):
    def __init__(self, ctr_id, rocis_count):
        self.ctr_id = ctr_id
        self.rocis_count = rocis_count
        self.bucket_overflow = 0
        self.approaching_overflow = 0

    def __repr__(self):
        return 'PerCtr(ctr_id={}, rocis_count={}, bucket_overflow={}, approaching_overflow={})'.format(self.ctr_id, self.rocis_count, self.bucket_overflow, self.approaching_overflow)

    def to_json(self):
        return {"ctr_id": self.ctr_id,
                "rocis_count": self.rocis_count,
                "bucket_overflow": self.bucket_overflow,
                "approaching_overflow": self.approaching_overflow}


class PerPStore(object):
    def __init__(self, total_rocis, device, per_ctrs):
        self.total_rocis = total_rocis
        self.device = device
        self.per_ctrs = per_ctrs

    def __repr__(self):
        return 'PerPStore(total_rocis={}, device={}, per_ctrs={})'.format(self.total_rocis, repr(self.device), repr(self.per_ctrs))

    def to_json(self):
        return {"total_rocis": self.total_rocis,
                "device": self.device,
                "per_ctrs": [obj.to_json() for obj in self.per_ctrs] if self.per_ctrs != None else None}


class PerEnode(object):
    def __init__(self, per_pstores, total_rocis):
        self.total_rocis = total_rocis
        self.per_pstores = per_pstores

    def __repr__(self):
        return 'PerEnode(total_rocis={}, per_pstores={})'.format(self.total_rocis, self.per_pstores)

    def to_json(self):
        return {"total_rocis": self.total_rocis,
                "per_pstores": [obj.to_json() for obj in self.per_pstores]}


class PerCluster(object):
    def __init__(self, per_enodes):
        self.per_enodes = per_enodes

    def __repr__(self):
        return 'PerCluster(per_enodes={})'.format(self.per_enodes)

    def to_json(self):
        return {"per_enodes:": [{'ip':key,"val":value.to_json()} for key,value in self.per_enodes.items()]}


class RocsMap(object):
    def __init__(self, bitshift, replication_level, per_cluster):
        self.bitshift = bitshift
        self.replication_level = replication_level
        self.per_cluster = per_cluster
    def __repr__(self):
        return 'RocsMap(bitshift={}, replication_level={}, per_cluster={})'.format(self.bitshift, self.replication_level, self.per_cluster)
    def to_json(self):
        return {"bitshift": self.bitshift,
                "replication_level": self.replication_level,
                "per_cluster:": self.per_cluster.to_json()}


def collect_roc_map(enodes):
    roc_set = cmd_output(['ecs-cli', 'roc-set'])

    '''
    ecs-cli output example:

    [root@dadee-cluster-b ~(elfs_admin)]# ecs-cli roc-set
    roc-replication: 2, bitshift: 12
    [45] vHead-10.0.1.78 10.0.1.78 EOid[ENODE:552ea8b9-e457-4be4-8259-9838d599658c:34] dstore:2 : 64
    [45] vHead-10.0.1.78 10.0.1.78 EOid[ENODE:552ea8b9-e457-4be4-8259-9838d599658c:34] dstore:1 : 64
    '''

    lines = roc_set.split("\n")
    caption_tokens = lines[0].split()
    assert caption_tokens[0] == "roc-replication:"
    replication_level = caption_tokens[1].strip(",")
    assert caption_tokens[2] == "bitshift:"
    bitshift = caption_tokens[3]

    per_enodes = {}

    for enode in enodes:

        per_pstores = []
        total_rocis_per_enode = 0

        for line in lines:
            if line.find("dstore:") == -1:
                continue
            enode_data_ip = line.split()[2]
            if enode_data_ip != enode.data_ip:
                continue
            total_rocis_per_pstore = int(line.split()[-1])
            total_rocis_per_enode += total_rocis_per_pstore
            per_pstores.append(PerPStore(total_rocis=total_rocis_per_pstore,
                                         device=None, per_ctrs=None))

        per_enodes[enode.data_ip] = PerEnode(total_rocis=total_rocis_per_enode,
                                             per_pstores=per_pstores)

    return RocsMap(bitshift=bitshift,
                   replication_level=replication_level,
                   per_cluster=PerCluster(per_enodes=per_enodes))


def ssh_to_output_file(target_ip, output_file_prefix, remote_cmd):
    filename = tempfile.mktemp(prefix='bo_monitoring.' + output_file_prefix + '.')
    with open(filename, 'w') as out_file:
        cmd_redirect_into(out_file, ['ssh'] + SSH_OPTS + [target_ip, remote_cmd])
    return filename

def collect_rocis_from_enode(output_file, enode):
    '''
    Output example of /elastifile/tools/show-counters -i 127.0.0.1 -p 8911 -u -l -s rocis_count

    searching counters in ecores: [0, 1, 2, 3, 4, 5, 6, 7, 8]
    data_store.944e34.roci_mapper.ctr_0.rocis_count (id=413, type=RO_INT, ecore=8): 1
    data_store.944e34.roci_mapper.ctr_1.rocis_count (id=414, type=RO_INT, ecore=8): 2
    data_store.944e34.roci_mapper.ctr_2.rocis_count (id=415, type=RO_INT, ecore=8): 1
    '''

    per_pstores = []
    total_rocis_per_enode = 0

    with open(output_file, 'r') as out_file:

        for device in enode.devices:
            per_ctrs = []
            total_rocis_per_pstore = 0

            out_file.seek(0)
            for line in out_file:
                if line.find(ROCIS_PER_CONTAINER_COUNTER_ID) == -1:
                  continue
                disk_id = line.split(".")[1]
                if device.find(disk_id) == -1:
                  continue
                found = True

                # TODO: check if output is different on master
                ctr_id = line.split(".")[3]
                count = int(line.split()[4])
                total_rocis_per_pstore += count
                per_ctrs.append(PerCtr(ctr_id=ctr_id, rocis_count=count))

            per_pstores.append(PerPStore(total_rocis=total_rocis_per_pstore,
                                         device=device,
                                         per_ctrs=per_ctrs))
            total_rocis_per_enode += total_rocis_per_pstore

    return PerEnode(total_rocis=total_rocis_per_enode,
                    per_pstores=per_pstores)

def find_ctr(disk_id, ctr_id, per_enode):
    for per_pstore in per_enode.per_pstores:
        if per_pstore.device.find(disk_id) == -1:
            continue

        for per_ctr in per_pstore.per_ctrs:
          if per_ctr.ctr_id != ctr_id:
              continue

          return per_ctr

    assert False


def read_bucket_overflow_counters(output_file, per_enode):
    '''
    Output example of /elastifile/tools/show-counters -i 127.0.0.1 -p 8911 -u -l -s direct.pdbl.save.bucket_

    searching counters in ecores: [0, 1, 2, 3, 4, 5, 6, 7, 8]
    data_store.944e34.pstore.ctr_0.direct.pdbl.save.bucket_overflowed (id=1943, type=RO_INT, ecore=4): 0
    data_store.944e34.pstore.ctr_0.direct.pdbl.save.bucket_approaching_overflow (id=1944, type=RO_INT, ecore=4): 5
    data_store.f6bd74.pstore.ctr_63.direct.pdbl.save.bucket_overflowed (id=6732, type=RO_INT, ecore=7): 1
    data_store.f6bd74.pstore.ctr_63.direct.pdbl.save.bucket_approaching_overflow (id=6733, type=RO_INT, ecore=7): 2456
    '''

    with open(output_file, 'r') as out_file:
        for line in out_file:
            if line.find("_overflow") == -1:
              continue
            count = int(line.split()[-1])
            if count == 0:
                continue

            tokens = line.split(".")
            (datastore, disk_id, pstore_str, ctr_id), rest = tokens[:4], tokens[4:]
            assert datastore == "data_store"
            assert pstore_str == "pstore"
            per_ctr = find_ctr(disk_id, ctr_id, per_enode)

            if line.find("bucket_overflowed") != -1:
                per_ctr.bucket_overflow = count
            else:
                per_ctr.approaching_overflow = count


def collect_cluster_data(enodes):
    per_enodes = {}
    print('Collecting counters from enodes...')
    for enode in enodes:
        output_file_prefix='enode_data'.format(enode.data_ip)

        roci_mapper_cmd="/elastifile/tools/show-counters -i 127.0.0.1 -p 8911 -u -l -s {}".format(ROCIS_PER_CONTAINER_COUNTER_ID)
        roci_mapper_output_file = ssh_to_output_file(enode.data_ip, output_file_prefix=output_file_prefix, remote_cmd=roci_mapper_cmd)
        per_enode = collect_rocis_from_enode(roci_mapper_output_file, enode)

        bucket_overflow_cmd="/elastifile/tools/show-counters -i 127.0.0.1 -p 8911 -u -l -s {}".format(BUCKET_OVERFLOW_COUNTER_ID)
        bucket_overflow_output_file = ssh_to_output_file(enode.data_ip, output_file_prefix=output_file_prefix, remote_cmd=bucket_overflow_cmd)
        read_bucket_overflow_counters(bucket_overflow_output_file, per_enode)

        per_enodes[enode.data_ip] = per_enode
        os.remove(roci_mapper_output_file)
        os.remove(bucket_overflow_output_file)

    return PerCluster(per_enodes=per_enodes)


def collect_bucket_overflow_traces(enodes):
    traces = {}
    print('Collecting traces...')
    for enode in enodes:
        try:
            output_file_prefix='dumper.{}.bucket_overflow'.format(enode.data_ip)
            remote_cmd="dumper.py -e /elastifile/bin/elfs --ut_files /dev/shm/elfs/elfs_ -c 6 -i '$fnc==\"check_bucket_overflow\"' -t 20"
            output_file = ssh_to_output_file(enode.data_ip, output_file_prefix=output_file_prefix, remote_cmd=remote_cmd)

            enode_traces = []
            with open(output_file, 'r') as out_file:
                for line in out_file:
                    if line.find("check_bucket_overflow") == -1:
                        continue
                    enode_traces.append(line)

            if len(enode_traces) > 0:
                traces[enode.data_ip] = enode_traces
        except:
            print "Ignoring failed dumper command for enode {}".format(enode.data_ip)
        finally:
            try:
                os.remove(output_file)
            except:
                print "Failed to remove {}".format(output_file)

    return traces

def get_statistics(enode_roci_mapper):

    histogram = [0] * MAX_ROCIS_COUNT_PER_CTR
    max_count = 0
    min_count = MAX_ROCIS_COUNT_PER_CTR
    bucket_overflow_ctrs = 0
    approaching_overflow_ctrs = 0

    for pstore in enode_roci_mapper.per_pstores:
        for ctr in pstore.per_ctrs:

            assert ctr.rocis_count < MAX_ROCIS_COUNT_PER_CTR
            histogram[ctr.rocis_count] += 1
            max_count = max(max_count, ctr.rocis_count)
            min_count = min (min_count, ctr.rocis_count)
            if ctr.bucket_overflow > 0:
                bucket_overflow_ctrs += 1
            if ctr.approaching_overflow > 0:
                approaching_overflow_ctrs += 1

    return (max_count, min_count, histogram, bucket_overflow_ctrs, approaching_overflow_ctrs)


def do_analysis(per_cluster, roc_map_data):

    cluster_histogram = [0] * MAX_ROCIS_COUNT_PER_CTR
    cluster_max_count = 0
    max_rocis_enodes = set()
    cluster_min_count = MAX_ROCIS_COUNT_PER_CTR

    for data_ip, enode_roc_map in roc_map_data.per_enodes.items():
        enode_roci_mapper = per_cluster.per_enodes[data_ip]

        if enode_roc_map.total_rocis != enode_roci_mapper.total_rocis:
            print "WARN_#unmapped_rocies: {} - Roc map differ from roci mapper ( {} != {} )".format(data_ip,
                                                                                                    enode_roc_map.total_rocis,
                                                                                                    enode_roci_mapper.total_rocis)

        max_count, min_count, histogram, bucket_overflow_ctrs, approaching_overflow_ctrs = get_statistics(enode_roci_mapper)

        if bucket_overflow_ctrs > 0:
            print "WARN_#bucket_overflow: {} - {} containers have/had overflowing buckets".format(data_ip,
                                                                                                  bucket_overflow_ctrs)
        if approaching_overflow_ctrs > 0:
            print "WARN_#bucket_overflow: {} - {} containers approach/ed overflow".format(data_ip,
                                                                                          approaching_overflow_ctrs)

        if max_count >= cluster_max_count:
            if max_count > cluster_max_count:
                cluster_max_count = max_count
                max_rocis_enodes.clear()
            max_rocis_enodes.add(data_ip)
        cluster_min_count = min (cluster_min_count, min_count)
        cluster_histogram = [sum(n) for n in zip(*[cluster_histogram, histogram])]

    if cluster_max_count - cluster_min_count >= MAX_ROCI_IMBALANCE_BETWEEN_CTRS:
        print "WARN_#roci_imbalance: max-{}, min-{}, histogram-{}".format(cluster_max_count,
                                                                          cluster_min_count,
                                                                          cluster_histogram)
        print "WARN_#roci_imbalance: nodes with max roci: {}".format(max_rocis_enodes)

    return (cluster_max_count, cluster_min_count, cluster_histogram, max_rocis_enodes)


def is_rebuild_running():
    roc_set = cmd_output(['ecs-cli', 'ha-report'])
    lines = roc_set.split("\n")

    for line in lines:
        if line.find("roc-transitions: total:") == -1:
            continue
        print line
        tokens = line.split()
        total_transitions = int(tokens[1].split(":")[1])
        done_transitions = int(tokens[2].split(":")[1])
        inprogress = int(tokens[3].split(":")[1])
        if total_transitions != done_transitions or inprogress > 0:
          return True
        return False

    return False


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--no-traces', action='store_true')
    args = parser.parse_args()

    if is_rebuild_running():
        sys.exit("ROC transitions are in progress. Please re-run after rebuild is done")

    print '----------------------------------------------------------------------'
    print 'Collecting data:'
    enodes = get_enodes()
    per_cluster = collect_cluster_data(enodes)
    roc_map = collect_roc_map(enodes)

    if not args.no_traces:
        traces = collect_bucket_overflow_traces(enodes)

    print '----------------------------------------------------------------------'
    print 'Findings:'
    max_count, min_count, histogram, max_rocis_enodes = do_analysis(per_cluster,
                                                                    roc_map.per_cluster)
    if not args.no_traces:
        if len(traces) > 0:
            print "WARN_#bucket_overflow: enodes with bucket_overflow traces: {}".format(traces.keys())

    print '----------------------------------------------------------------------'
    print 'Detailed information:'
    print '-------- histogram -------------'
    print "max: {}, min: {}, histogram: {}".format(max_count,
                                                   min_count,
                                                   histogram)
    print "enodes with max rocis: {}".format(max_rocis_enodes)

    if not args.no_traces:
        print '-------- traces -------------'
        for data_ip, enode_traces in traces.items():
            print "-------------> {}".format(data_ip)
            print enode_traces

    print '--------- roc map ------------'
    roc_map_json = json.dumps(roc_map.to_json(), indent=2)
    print roc_map_json
    print '--------- roci mapper ------------'
    per_cluster_json = json.dumps(per_cluster.to_json(), indent=2)
    print per_cluster_json
    print '----------------------------------------------------------------------'


if __name__=='__main__':
    main()
